package de.uniwue.mk.kall.athen.luceneView.widget;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.uima.cas.Feature;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.TypeSystem;
import org.eclipse.jface.viewers.CheckboxTreeViewer;
import org.eclipse.jface.viewers.ITreeViewerListener;
import org.eclipse.jface.viewers.TreeExpansionEvent;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerComparator;
import org.eclipse.jface.viewers.ViewerFilter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.TreeItem;

public class TypeBrowserWidget extends Composite {
	private TreeViewer viewer;
	private ViewerFilter typeNameFilter;
	private String typeNamePart = "";
	/**
	 * This set is used to store information about which types and features 
	 * were checked (especially those that are no longer displayed in the 
	 * TreeViewer because of a filter; the TreeViewer 'forgets' their 
	 * status). 
	 */
	private Set<Object> checkedItems;


	public TypeBrowserWidget(Composite parent, boolean displayCheckboxes) {
		super(parent, SWT.BORDER);
		checkedItems = new HashSet<>();
		createControls(displayCheckboxes);
	}


	public void createControls(boolean displayCheckboxes) {
		setLayout(new GridLayout(2, false));

		/*
		 * Label + Text for filtering the tree
		 */
		Label lbFilter = new Label(this, SWT.NONE);
		lbFilter.setText("Filter by Name:");

		Text tfFilter = new Text(this, SWT.NONE);
		tfFilter.setMessage("Enter Name to filter by");

		GridData gdTfFilter = new GridData(GridData.FILL_HORIZONTAL);
		gdTfFilter.horizontalSpan = 1;
		gdTfFilter.grabExcessHorizontalSpace = true;
		tfFilter.setLayoutData(gdTfFilter);

		tfFilter.addModifyListener(event -> {
			typeNamePart = tfFilter.getText();
			viewer.refresh(false);
			TreeItem[] items = viewer.getTree().getItems();
			//			new Thread(() -> checkStatus(items, 0, items.length)).run();
			checkStatus(items, 0, items.length);
		});

		/*
		 * TreeViewer
		 */
		if (displayCheckboxes) {
			viewer = new CheckboxTreeViewer(this, SWT.H_SCROLL | SWT.V_SCROLL);
			createTreeListener();
			createCheckListener();
		} else {
			viewer = new TreeViewer(this, SWT.H_SCROLL | SWT.V_SCROLL);
		}

		GridData gdViewer = new GridData(GridData.FILL_BOTH);
		gdViewer.heightHint = 300;
		gdViewer.horizontalSpan = 2;
		viewer.getTree().setLayoutData(gdViewer);

		viewer.setContentProvider(new TypeBrowserContentProvider());
		viewer.setLabelProvider(new TypeBrowserLabelProvider());

		createTypeNameFilter();
		viewer.addFilter(typeNameFilter);
		viewer.setComparator(new ViewerComparator());
		viewer.getTree().setSortDirection(SWT.UP);

	}


	private void createTypeNameFilter() {
		typeNameFilter = new ViewerFilter() {

			@Override
			public boolean select(Viewer viewer, Object parentElement, Object element) {
				/*
				 *  only show Types if their name contains typeNamePart
				 *  toLowerCase => ignore case
				 */
				return !(element instanceof Type) || 
						((Type) element).getName().toLowerCase().contains(
								typeNamePart.toLowerCase());
			}

		};
	}

	/**
	 * Creates a listener that reacts to expansion events in the tree. When an 
	 * element is expanded all its children that are in checkedItems will be set 
	 * to checked (doing this in {@link #checkStatus(TreeItem[], int, int)} does 
	 * not work (widget disposed)). 
	 */
	private void createTreeListener() {
		viewer.addTreeListener(new ITreeViewerListener() {

			@Override
			public void treeExpanded(TreeExpansionEvent event) {
				Object element = event.getElement();
				checkStatusOfChildren(element);
			}

			@Override
			public void treeCollapsed(TreeExpansionEvent event) {
			}
		});
	}

	private void createCheckListener() {
		viewer.getTree().addSelectionListener(new SelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				if (e.item instanceof TreeItem && e.detail == SWT.CHECK) {
					/*
					 * propagate changes to parent / children
					 * one or more Features checked => Type checked
					 */
					TreeItem item = (TreeItem) e.item;
					if (item.getData() instanceof Type && !(item.getChecked())) {
						// type unchecked => uncheck features
						for (TreeItem child : item.getItems()) {
							child.setChecked(item.getChecked());
							checkedItems.remove(child.getData());
						}
					} else if (item.getData() instanceof Feature && item.getChecked()) {
						// feature checked => check type
						item.getParentItem().setChecked(true);
						checkedItems.add(item.getParentItem().getData());
					}
					/*
					 * propagate change to checkedItems
					 */
					if (item.getChecked()) {
						checkedItems.add(item.getData());
					} else {
						checkedItems.remove(item.getData());
					}
				}
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
			}
		});
	}


	/**
	 * Set all given TreeItems between index <i>start</i> (inclusive) and index 
	 * <i>stop</i> (exclusive) whose data object is contained in <i>checkedItems</i> 
	 * to checked. </br>
	 * This method is used after the TreeViewers filter changes since the TreeViewer 
	 * loses information about whether an item that is no longer displayed because 
	 * of a filter was checked or not. 
	 * @param items
	 * @param start
	 * @param stop
	 */
	private void checkStatus(TreeItem[] items, int start, int stop) {
		if (start < 0) {
			start = 0;
		}
		if (stop > items.length) {
			stop = items.length;
		}
		for (int i = start; i < stop; i++) {
			if (checkedItems.contains(items[i].getData())) {
				TreeItem item = items[i];
				item.setChecked(true);
			}
		}
	}

	private void checkStatusOfChildren(Object element) {
		TreeItem item = null;
		for (TreeItem ti : viewer.getTree().getItems()) {
			if (element.equals(ti.getData())) {
				item = ti;
				break;
			}
		}
		if (item == null || !item.getChecked()) {
			return;
		}
		for (TreeItem child : item.getItems()) {
			if (checkedItems.contains(child.getData())) {
				child.setChecked(true);
			}
		}
	}


	public TreeViewer getViewer() {
		return viewer;
	}

	public void setTypeSystem(TypeSystem system) {
		viewer.setInput(system);
	}

	/**
	 * Collects all Types and corresponding Features that were selected in the TreeViewer 
	 * in a Map.
	 * @return
	 */
	public Map<Type, List<Feature>> getSelectedElements() {
		HashMap<Type, List<Feature>> map = new HashMap<>();
		for (Object obj : checkedItems) {
			// Only process types. Features will be processed with their type. 
			if (!(obj instanceof Type)) {
				continue;
			}
			Type type = (Type) obj;
			List<Feature> features = collectSelectedFeatures(type);
			map.put(type, features);
		}
		return map;
	}

	/**
	 * Collects all the Features belonging to this Type that are checked.
	 * @param items
	 * @return
	 */
	private List<Feature> collectSelectedFeatures(Type type) {
		List<Feature> list = new ArrayList<>();
		for (Feature feat : type.getFeatures()) {
			if (checkedItems.contains(feat)) {
				list.add(feat);
			}
		}
		return list;
	}
}
