package de.uniwue.mk.kall.lucene;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;

import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.FieldType;
import org.apache.lucene.document.StoredField;
import org.apache.lucene.index.DirectoryReader;
import org.apache.lucene.index.IndexOptions;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.IndexWriterConfig;
import org.apache.lucene.queryparser.classic.ParseException;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.PhraseQuery;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.store.FSDirectory;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.cas.impl.XmiCasDeserializer;
import org.apache.uima.cas.text.AnnotationFS;
import org.xml.sax.SAXException;

import de.uniwue.mkrug.kall.typesystemutil.Util_impl;

public class AthenLuceneIndex {
	public static final String FIELDNAME_TEXT = "Text";
	public static final String FIELDNAME_FILE = "File";

	private File directory;
	private IndexTypeConfiguration config;
	private CAS cas;


	private AthenLuceneIndex() {
		cas = Util_impl.createCas();
	}


	public File getDirectory() {
		return directory;
	}

	public IndexTypeConfiguration getConfiguration() {
		return config;
	}

	public void addDocuments(File directoryDocs) throws IOException {
		if (!directoryDocs.isDirectory()) {
			throw new IllegalArgumentException("The file object needs to be a directory!");
		}

		IndexWriter writer = openWriter();

		for (File file : directoryDocs.listFiles(fo -> fo.isFile() && fo.getName().endsWith("xmi"))) {
			addDocument(writer, file);
		}

		closeWriter(writer);
	}

	public void closeWriter(IndexWriter writer) throws IOException {
		writer.close();
		writer.getDirectory().close();
	}

	public IndexWriter openWriter() throws IOException {
		FSDirectory fsdir = FSDirectory.open(directory.toPath());
		AthenLuceneAnalyzer analyzer = new AthenLuceneAnalyzer(config);
		IndexWriterConfig iwCfg = new IndexWriterConfig(analyzer);
		return new IndexWriter(fsdir, iwCfg);
	}

	public void addDocument(IndexWriter writer, File file) {
		AthenLuceneAnalyzer analyzer = (AthenLuceneAnalyzer) writer.getAnalyzer();
		try {
			XmiCasDeserializer.deserialize(new FileInputStream(file), cas);
		} catch (IOException | SAXException e) {
			e.printStackTrace();
			return;
		}
		analyzer.setCas(cas);
		for (AnnotationFS anno : cas.getAnnotationIndex(config.getDocumentPartType())) {
			Document doc = new Document();
			analyzer.setDocumentPart(anno);
			FieldType type = new FieldType();
			type.setTokenized(true);
			type.setStoreTermVectors(true);
			type.setStoreTermVectorPositions(true);
			type.setStored(true);
			type.setIndexOptions(IndexOptions.DOCS_AND_FREQS_AND_POSITIONS);
			Field fieldText = new Field(FIELDNAME_TEXT, anno.getCoveredText(), type);
			doc.add(fieldText);
			doc.add(new StoredField(FIELDNAME_FILE, file.getAbsolutePath()));
			try {
				writer.addDocument(doc);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	public ArrayList<AthenLuceneResult> query(String queryText) throws IOException, ParseException {
		FSDirectory fsdir = FSDirectory.open(directory.toPath());
		DirectoryReader reader = DirectoryReader.open(fsdir);
		IndexSearcher searcher = new IndexSearcher(reader);

		PhraseQuery q = new PhraseQuery(0, "Text", queryText.split(" "));

		ScoreDoc[] scoreDocs = searcher.search(q, 1000).scoreDocs;
		ArrayList<AthenLuceneResult> list = new ArrayList<>();

		for (ScoreDoc doc : scoreDocs) {
			int idx = doc.doc;
			String text = searcher.doc(idx).getField(FIELDNAME_TEXT).stringValue();
			String path = searcher.doc(idx).getField(FIELDNAME_FILE).stringValue();
			list.add(new AthenLuceneResult(path, text));
		}

		reader.close();
		fsdir.close();
		return list;
	}


	public static AthenLuceneIndex createIndex(File directory, IndexTypeConfiguration config) 
			throws IOException {
		if (!directory.exists()) {
			directory.mkdir();
		}
		if (!directory.isDirectory()) {
			throw new IllegalArgumentException("the given file object has to be a directory!");
		}
		config.storeInDirectory(directory);
		AthenLuceneIndex index = new AthenLuceneIndex();
		index.directory = directory;
		index.config = config;
		return index;
	}

	public static AthenLuceneIndex loadIndex(File directory, TypeSystem ts) throws IOException {
		if (!directory.exists() || !directory.isDirectory()) {
			throw new IllegalArgumentException("the given file object has to be a directory!");
		}
		AthenLuceneIndex index = new AthenLuceneIndex();
		index.directory = directory;
		index.config = IndexTypeConfiguration.loadFromDirectory(directory, ts);
		return index;
	}
}
