package de.uniwue.mk.kall.lucene;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.lucene.analysis.Tokenizer;
import org.apache.lucene.analysis.tokenattributes.CharTermAttribute;
import org.apache.lucene.analysis.tokenattributes.PositionIncrementAttribute;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.Feature;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.text.AnnotationFS;

import de.uniwue.mkrug.kall.typesystemutil.Util_impl;

public class AthenLuceneTokenizer extends Tokenizer {
	private AnnotationFS documentPart;
	private Util_impl util;
	private IndexTypeConfiguration config;
	private List<AnnotationFS> tokenList;
	private List<Type> typeList;

	private PositionIncrementAttribute posIncrement;
	private CharTermAttribute termAttribute;

	private int tokenIndex;
	private int typeIndex;
	private int featureIndex;
	/**
	 * Indicates whether the token index has changed and {@link #posIncrement} 
	 * should be set to 1. </br>
	 * This is necessary since not all types can be indexed on all tokens. In such 
	 * a case where a type is not found on the current token the type index is 
	 * incremented and {@link #incrementToken()} is called. At this point the type 
	 * index does no longer indicate a position increment. 
	 */
	private boolean tokenHasChanged;


	public AthenLuceneTokenizer(IndexTypeConfiguration configuration) {
		posIncrement = addAttribute(PositionIncrementAttribute.class);
		termAttribute = addAttribute(CharTermAttribute.class);
		config = configuration;
		tokenIndex = 0;
		typeIndex = 0;
		featureIndex = 0;
		tokenHasChanged = true;
		typeList = new ArrayList<>(configuration.getTypesAndFeatures().keySet());
	}


	public void setCas(CAS cas) {
		util = new Util_impl(cas);
	}

	public void setDocumentPart(AnnotationFS docPart) {
		documentPart = docPart;
		tokenIndex = 0;
		typeIndex = 0;
		tokenList = util.getCovered(documentPart, config.getTokenType());
		tokenHasChanged = true;
	}


	@Override
	public boolean incrementToken() throws IOException {
		if (tokenIndex >= tokenList.size()) {
			return false;
		}

		clearAttributes();

		AnnotationFS token = tokenList.get(tokenIndex);

		/*
		 * Indexing cases:
		 * 	1.1	feature name = feature value
		 *  1.2	type name
		 *  2	plain text
		 */
		String value = "";
		if (typeIndex < typeList.size()) {
			/*
			 * Index all types and their features.
			 */
			Type type = typeList.get(typeIndex);
			List<AnnotationFS> covered = util.getCovered(token, type);
			if (covered.isEmpty()) {
				incrementType();
				return incrementToken();
			}
			if (featureIndex < config.getTypesAndFeatures().get(type).size()) {
				/*
				 * Index feature
				 */
				AnnotationFS anno = covered.get(0);
				Feature feature = config.getTypesAndFeatures().get(type).get(featureIndex);
				value = feature.getShortName() + "=" + anno.getFeatureValueAsString(feature);
			} else {
				/*
				 * Index type by name
				 */
				value = type.getShortName();
			}
		} else {
			/*
			 * Index the plain text.
			 */
			value = token.getCoveredText();
		}

		char[] charArray = value.toCharArray();
		termAttribute.copyBuffer(charArray, 0, charArray.length);

		if (tokenHasChanged) {
			posIncrement.setPositionIncrement(1);
			tokenHasChanged = false;
		} else {
			posIncrement.setPositionIncrement(0);
		}

		incrementFeature();

		return true;
	}

	/**
	 * Increments the feature index. 
	 * The last feature of each type is the type name (not actually in a list; 
	 * index = size of feature list). 
	 * If all features of the current type have been indexed the type index will 
	 * be incremented. 
	 */
	private void incrementFeature() {
		featureIndex++;
		/*
		 * plain text is no real type and therefore has no features
		 */
		if (typeIndex >= typeList.size()) {
			incrementType();
			return;
		}
		Type type = typeList.get(typeIndex);
		if (featureIndex > config.getTypesAndFeatures().get(type).size()) {
			incrementType();
		}
	}

	/**
	 * Increments the type index. 
	 * The last type is the plain text (not actually in a list; 
	 * index = size of type list). 
	 * If all types have been indexed the token index will be incremented and 
	 * {@link #tokenHasChanged} will be set to true. 
	 */
	private void incrementType() {
		featureIndex = 0;
		typeIndex++;
		if (typeIndex > typeList.size()) {
			typeIndex = 0;
			tokenIndex++;
			tokenHasChanged = true;
		}
	}

}
