package de.uniwue.mk.kall.lucene;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.uima.cas.Feature;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.TypeSystem;

public class IndexTypeConfiguration {
	public static final String DEFAULTFILENAME = "types.cfg";
	private static final String typeSeparator = "\n";
	private static final String featureSeparator = "\t";

	private Map<Type, List<Feature>> types;
	private Type tokenType;
	private Type documentPartType;


	public IndexTypeConfiguration(Map<Type, List<Feature>> typesAndFeatures, Type tokenType, 
			Type documentPartType) {
		if (typesAndFeatures == null || tokenType == null || documentPartType == null) {
			throw new IllegalArgumentException(
					"types, documentparttype and tokentype may not be null!");
		}
		if (!typesAndFeatures.keySet().contains(tokenType)) {
			typesAndFeatures.put(tokenType, new ArrayList<>());
		}
		this.types = typesAndFeatures;
		this.tokenType = tokenType;
		this.documentPartType = documentPartType;
	}


	public Map<Type, List<Feature>> getTypesAndFeatures() {
		return types;
	}

	public Type getTokenType() {
		return tokenType;
	}

	public Type getDocumentPartType() {
		return documentPartType;
	}

	public void storeInDirectory(File directory) throws IOException {
		storeInFile(new File(directory, DEFAULTFILENAME));
	}

	public void storeInFile(File file) throws IOException {
		/*
		 * Store the tokentype in the first line. 
		 * After that, store each type and its features in one line. 
		 * All types and features are stored via their full names. 
		 */
		FileWriter writer = new FileWriter(file);
		writer.write(tokenType.getName() + typeSeparator);
		writer.write(documentPartType.getName() + typeSeparator);
		for (Type type : types.keySet()) {
			ArrayList<String> list = new ArrayList<>();
			list.add(type.getName());
			types.get(type).forEach(feature -> list.add(feature.getName()));
			writer.write(String.join(featureSeparator, list) + typeSeparator);
		}
		writer.flush();
		writer.close();
	}

	public static IndexTypeConfiguration loadFromDirectory(File directory, TypeSystem typeSystem) 
			throws IOException {
		return loadFromFile(new File(directory, DEFAULTFILENAME), typeSystem);
	}

	public static IndexTypeConfiguration loadFromFile(File file, TypeSystem typeSystem) 
			throws IOException {
		/*
		 * The first line contains the name of the tokentype. 
		 * Each line after that contains the name of a type and the names of its features. 
		 * Since the file only contains the names of types and features, the typeSystem is 
		 * needed to get the actual types and features. 
		 */
		BufferedReader reader = new BufferedReader(new FileReader(file));
		String line = reader.readLine();
		Type tokenType = typeSystem.getType(line);
		if (tokenType == null) {
			// error (throw exception instead?)
			reader.close();
			return null;
		}
		line = reader.readLine();
		Type documentPartType = typeSystem.getType(line);
		if (documentPartType == null) {
			// error (throw exception instead?)
			reader.close();
			return null;
		}
		Map<Type,List<Feature>> map = new HashMap<>();
		while ((line = reader.readLine()) != null) {
			String[] split = line.split(featureSeparator);
			Type type = typeSystem.getType(split[0]);
			if (type == null) {
				continue;
			}
			ArrayList<Feature> set = new ArrayList<>();
			for (int i = 1; i < split.length; i++) {
				set.add(typeSystem.getFeatureByFullName(split[i]));
			}
			map.put(type, set);
		}
		reader.close();
		return new IndexTypeConfiguration(map, tokenType, documentPartType);
	}
}
