package de.uniwue.mk.medIE.extraction.algorithm;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.uima.cas.CAS;
import org.apache.uima.cas.Feature;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.text.AnnotationFS;
import org.apache.uima.cas.text.AnnotationIndex;

import de.uniwue.mk.medIE.terminology.Concept;
import de.uniwue.mk.medIE.terminology.ENodeType;
import de.uniwue.mk.medIE.terminology.LexikonEntry;
import de.uniwue.mk.medIE.terminology.MedicalTerminology;
import de.uniwue.mk.medIE.terminology.Variation;
import de.uniwue.mkrug.kall.typesystemutil.Util_impl;

public class SimpleInSegmentMatchingAlgorithm extends ATerminologyMatchingAlgorithm {

	private static final String REGEX_LARGER = "((<|>|<=|>=|=)";
	private static final String REGEX_NUMBER = "[0-9]+[\\.,]?([0-9]+)?";
	private static final String REGEX_NUMBER_MATCH = REGEX_NUMBER + "([-]" + REGEX_NUMBER + ")?";
	private Type conceptType;
	private Feature nodeType;
	private Feature typeContent;
	private Feature relatedTo;
	private Feature conceptId;

	@Override
	public void apply(MedicalTerminology terminology, CAS cas) {

		if (terminology == null)
			return;

		conceptType = cas.getTypeSystem().getType("de.uniwue.medIE.Concept");

		// Structure /Attribute/Value ...
		nodeType = conceptType.getFeatureByBaseName("NodeType");
		// BOOL, BOOL PLUS;CHOICE etc...
		typeContent = conceptType.getFeatureByBaseName("Type");
		relatedTo = conceptType.getFeatureByBaseName("RelatedTo");
		conceptId = conceptType.getFeatureByBaseName("ConceptID");

		// STEP 1 get all attributes of the terminology
		List<Concept> attributeNodes = terminology.getAttributeNodes();

		Util_impl util = new Util_impl(cas);

		AnnotationIndex<AnnotationFS> sentences = cas.getAnnotationIndex(util.getSentenceType());

		initProgress(attributeNodes.size() * sentences.size() + 1);

		// for each Attribute - match it to the text
		for (AnnotationFS sent : sentences) {

			for (Concept attribute : attributeNodes) {
				increaseProgress(1);

				ENodeType attType = attribute.getNodeType();

				// only apply for those kind of segments
				if (!(attType == ENodeType.CHOICE || attType == ENodeType.NUMERIC || attType == ENodeType.UNIT))
					continue;

				// check all segments - stored in a sentence annotation
				AnnotationFS att = matchToSegment(attribute, sent, cas);

				if (att == null)
					continue;

				// now theres an attribute - treat it accordingly to its type

				String type = att.getFeatureValueAsString(typeContent);

				// handle the CHOICE
				if (type.equals(ENodeType.CHOICE.toString())) {

					extractChoice(terminology, cas, attribute, sent, att);

				}

				else if (type.equals(ENodeType.NUMERIC.toString())) {

					extractNumeric(terminology, cas, attribute, sent, att);

				}

				// same as numeric but with a UNIT
				else if (type.equals(ENodeType.UNIT.toString())) {

					extractUnits(terminology, cas, attribute, sent, att);
				}

			}
		}

		// update the progress label

	}

	private void extractUnits(MedicalTerminology terminology, CAS cas, Concept attribute, AnnotationFS sent,
			AnnotationFS att) {
		// we assume the value are all of the type >|< XX
		List<Concept> values = terminology.getValuesToConcept(attribute);

		if (values.size() == 0) {

			// we search for nothing but the number itself

			boolean foundValue = false;
			List<String> units = attribute.getUnits();

			// TODO are multiple matches allowed ???
			for (String unit : units) {
				Matcher matcher = Pattern.compile(
						REGEX_LARGER + "(\\s*))?" + REGEX_NUMBER_MATCH + "(\\s*)" + Pattern.quote(unit)).matcher(
						sent.getCoveredText());

				while (matcher.find()) {

					if (matcher.end() + sent.getBegin() < att.getBegin())
						continue;

					// only accept values after the concept
					AnnotationFS anno = cas.createAnnotation(conceptType, sent.getBegin() + matcher.start(),
							sent.getBegin() + matcher.end());

					// add the features
					anno.setFeatureValueFromString(conceptId, String.valueOf(attribute.getId()));
					anno.setFeatureValueFromString(nodeType, "VALUE");
					anno.setFeatureValueFromString(typeContent, "NUMERIC");
					foundValue = true;

					cas.addFsToIndexes(anno);
					// the value was found in brackets?

					Matcher matcherBrackets = Pattern.compile("\\(.*?" + Pattern.quote(matcher.group()) + ".*?\\)")
							.matcher(sent.getCoveredText());

					while (matcherBrackets.find()) {
						// no match
						foundValue = false;
						cas.removeFsFromIndexes(anno);
						break;

					}

					break;
				}
			}

			if (!foundValue) {
				cas.removeFsFromIndexes(att);
			}
		}

		else {

			// we assume that we have to use one of the given
			// concepts TODO
		}
	}

	private void extractNumeric(MedicalTerminology terminology, CAS cas, Concept attribute, AnnotationFS sent,
			AnnotationFS att) {
		// we assume the value are all of the type >|< XX
		List<Concept> values = terminology.getValuesToConcept(attribute);

		if (values.size() == 0) {

			// we search for nothing but the number itself

			Matcher matcher = Pattern.compile(REGEX_LARGER + "(\\s*))?" + REGEX_NUMBER_MATCH).matcher(
					sent.getCoveredText());

			while (matcher.find()) {

				if (matcher.end() + sent.getBegin() < att.getBegin())
					continue;

				// only accept values after the concept
				AnnotationFS anno = cas.createAnnotation(conceptType, sent.getBegin() + matcher.start(),
						sent.getBegin() + matcher.end());

				// add the features
				anno.setFeatureValueFromString(conceptId, String.valueOf(attribute.getId()));
				anno.setFeatureValueFromString(nodeType, "VALUE");
				anno.setFeatureValueFromString(typeContent, "NUMERIC");

				cas.addFsToIndexes(anno);

				break;
			}

		}

		else {

			// we assume that we have to use one of the given
			// concepts TODO
		}
	}

	private void extractChoice(MedicalTerminology terminology, CAS cas, Concept attribute, AnnotationFS sent,
			AnnotationFS att) {
		// exactly 1 value has to be extracted or none and we can
		// still keep the annotation

		List<Concept> values = terminology.getValuesToConcept(attribute);
		// we can extract "vorhanden" if it is not negated
		boolean isNegated = checkForNegation(cas, new Util_impl(cas), att);

		List<AnnotationFS> matchedValues = new ArrayList<AnnotationFS>();
		for (Concept value : values) {

			AnnotationFS val = matchValueToAttribute(attribute, att, value, sent, cas);

			if (val != null) {
				matchedValues.add(val);
			}

		}

		if (matchedValues.size() == 0 && !attribute.isPureExistence()) {

			// remove from index again
			cas.removeFsFromIndexes(att);
		}

		if (matchedValues.size() == 0 && attribute.isPureExistence()) {

			if (isNegated) {
				att.setFeatureValueFromString(typeContent, "CHOICE_NICHT_VORHANDEN");
			}

			else {
				att.setFeatureValueFromString(typeContent, "CHOICE_VORHANDEN");
				att.setFeatureValueFromString(nodeType, "ATTRIBUTE_VORHANDEN");
			}

		}

		// if we have exactly one match we use this one
		else if (matchedValues.size() == 1) {

			AnnotationFS val = matchedValues.get(0);

			cas.addFsToIndexes(val);

		}

		else if (matchedValues.size() > 1) {

			// we use only the largest match

			AnnotationFS largest = null;

			for (AnnotationFS a : matchedValues) {

				if (largest == null)
					largest = a;

				else {

					int lenA = a.getEnd() - a.getBegin();
					int lenLargest = largest.getEnd() - largest.getBegin();

					if (lenA > lenLargest) {
						largest = a;
					}
				}
			}

			// add largest to CAS
			cas.addFsToIndexes(largest);
		}

	}

	private AnnotationFS matchValueToAttribute(Concept attribute, AnnotationFS att, Concept value, AnnotationFS sent,
			CAS cas) {
		List<String> synonymsAsRegex = value.getSynonymsAsRegex();
		for (String s : synonymsAsRegex) {

			Matcher m = Pattern.compile(s).matcher(sent.getCoveredText());

			while (m.find()) {

				AnnotationFS valAnno = cas.createAnnotation(conceptType, sent.getBegin() + m.start(), sent.getBegin()
						+ m.end());

				// add the features
				valAnno.setFeatureValueFromString(conceptId, String.valueOf(value.getId()));
				valAnno.setFeatureValueFromString(nodeType, "VALUE");
				valAnno.setFeatureValueFromString(typeContent, value.getNodeType().toString());

				valAnno.setFeatureValue(relatedTo, att);

				att.setFeatureValue(relatedTo, valAnno);

				// add only the biggest match

				return valAnno;
				// we can only apply each attribute once
			}
		}

		return null;
	}

	private void extractBool(CAS cas, Util_impl util, AnnotationFS att) {
		// we extract in any case just the attribute - but not the
		// value - since we dont need any

		// check if this is is not negated
		boolean isNegated = checkForNegation(cas, util, att);

		if (isNegated) {
			att.setFeatureValueFromString(typeContent, "BOOL_NICHT_VORHANDEN");
		}

		else {
			att.setFeatureValueFromString(typeContent, "BOOL_VORHANDEN");
		}
	}

	private boolean checkForNegation(CAS cas, Util_impl util, AnnotationFS att) {

		List<AnnotationFS> covered = util.getCovered(att, util.getPOSType());

		if (covered.size() == 0)
			return false;

		AnnotationFS previousAnnotation = util.getPreviousAnnotation(covered.get(0));

		if (previousAnnotation == null)
			return false;

		if (previousAnnotation.getCoveredText().toLowerCase().matches("nicht|kein|keine|keiner|keines|keinen")) {
			return true;
		}

		return false;
	}

	private AnnotationFS matchToSegment(Concept attribute, AnnotationFS segment, CAS cas) {

		// use all synonyms and match those to the text

		List<String> synonymsAsRegex = attribute.getSynonymsAsRegex();
		List<AnnotationFS> matches = new ArrayList<AnnotationFS>();
		for (String s : synonymsAsRegex) {

			Matcher m = Pattern.compile(s).matcher(segment.getCoveredText());

			while (m.find()) {

				AnnotationFS attAnno = cas.createAnnotation(conceptType, segment.getBegin() + m.start(),
						segment.getBegin() + m.end());

				// add the features
				attAnno.setFeatureValueFromString(conceptId, String.valueOf(attribute.getId()));
				attAnno.setFeatureValueFromString(nodeType, "ATTRIBUTE");
				attAnno.setFeatureValueFromString(typeContent, attribute.getNodeType().toString());

				matches.add(attAnno);
				// cas.addFsToIndexes(attAnno);

				// we can only apply each attribute once
			}
		}

		AnnotationFS largest = null;

		for (AnnotationFS a : matches) {

			if (largest == null)
				largest = a;

			else {

				int lenA = a.getEnd() - a.getBegin();
				int lenLargest = largest.getEnd() - largest.getBegin();

				if (lenA > lenLargest) {
					largest = a;
				}
			}
		}

		if (largest == null)
			return null;
		// now respect the context of the annotation
		LexikonEntry context = attribute.getContext();

		// add largest to CAS
		if (largest != null && context == null) {
			cas.addFsToIndexes(largest);
		}

		if (context == null)
			return largest;

		// if we have a context we need to check every variation and see whether
		// the segment has that variation
		boolean hasContext = checkSegmentForContext(context, segment);

		if (!hasContext)
			return null;

		// return the match if the context was found
		cas.addFsToIndexes(largest);
		return largest;
	}

	private boolean checkSegmentForContext(LexikonEntry context, AnnotationFS segment) {

		if (context == null || segment == null)
			return false;

		for (Variation v : context.getAllVariations()) {

			for (String s : v.getVariationsAsList()) {

				Matcher m = Pattern.compile(s).matcher(segment.getCoveredText());

				// first goal wins
				while (m.find()) {
					return true;
				}
			}
		}

		return false;
	}
}
