package de.uniwue.mk.medIE.terminology;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import javax.xml.bind.annotation.XmlTransient;

public class Concept implements IGotVariations {

	private static final String PURE_EXISTENCE = "pure existence";

	private static final String IS_VALUE = "Is Value";

	private static final String IS_STRUCTURE = "is Structure";

	private static final String IS_OBJECT_ATTRIBUTE = " is ObjectAttribute";

	private static final String IS_IMPLICIT = "Is Implicit";

	private static final String IS_ATTRIBUTE = "Is Attribute";

	private String name;

	// TODO zeitstempel
	private LexikonEntry context;

	private Set<Variation> synonyms;

	@XmlTransient
	private Concept parent;

	private List<Concept> children;

	private HashMap<String, Object> properties;

	private boolean isRoot;

	private boolean userSetAttribute;
	private boolean isAttribute;
	private boolean isValue;
	private boolean userSetValue;

	private ENodeType nodeType;

	private boolean isImplicit;

	private boolean isObjectAttribute;

	// negated in the name - otherwise would be confusing
	private boolean isPureExistence;

	private boolean isStructure;

	// Dictionaries
	private List<LexikonEntry> references;

	private int id;

	public Concept() {

		this.synonyms = new HashSet<Variation>();
		this.properties = new HashMap<String, Object>();
		this.references = new ArrayList<LexikonEntry>();
	}

	public Concept(String name, Concept parent, List<Concept> children, int id) {
		super();
		this.name = name;
		this.parent = parent;
		this.children = children;
		this.id = id;

		this.synonyms = new HashSet<Variation>();
		this.properties = new HashMap<String, Object>();

		this.references = new ArrayList<LexikonEntry>();
	}

	public Concept(String name, Set<Variation> synonyms, Concept parent, List<Concept> children, int id) {
		super();
		this.name = name;
		this.synonyms = synonyms;
		this.parent = parent;
		this.children = children;
		this.id = id;

		this.properties = new HashMap<String, Object>();
		this.references = new ArrayList<LexikonEntry>();
	}

	/**
	 * Constructor that clones the given concept and all its children (using the
	 * same constructor to clone them). <br/>
	 * Note: IDs of the concept and its descendants will NOT be set, this will
	 * have to be done separately if necessary.
	 * 
	 * @param other
	 *            - the concept that will be cloned
	 * @param terminology
	 */
	public Concept(Concept other) {

		this.name = other.name;
		this.nodeType = other.nodeType;
		// copy boolean
		this.isRoot = other.isRoot;
		this.userSetAttribute = other.userSetAttribute;
		this.isAttribute = other.isAttribute;
		this.isValue = other.isValue;
		this.userSetValue = other.userSetValue;
		this.isImplicit = other.isImplicit;
		this.isObjectAttribute = other.isObjectAttribute;
		this.isPureExistence = other.isPureExistence;
		this.isStructure = other.isStructure;

		this.context = other.context;

		this.synonyms = other.synonyms.parallelStream().map(variation -> new Variation(variation))
				.collect(Collectors.toSet());

		this.synonyms = new HashSet<Variation>();
		for (Variation variation : other.synonyms) {
			this.synonyms.add(new Variation(variation));
		}

		this.parent = other.parent;

		this.children = other.children.stream().map(child -> new Concept(child)).collect(Collectors.toList());

		this.properties = new HashMap<>();
		other.properties.entrySet().parallelStream()
				.forEach(entry -> this.properties.put(entry.getKey(), entry.getValue()));

		this.children = new ArrayList<Concept>();
		for (Concept child : other.children) {
			this.children.add(new Concept(child));
		}

		this.properties = new HashMap<String, Object>();
		for (String key : other.properties.keySet()) {
			this.properties.put(key, other.properties.get(key));
		}
		this.children = new ArrayList<Concept>();
		for (Concept child : other.children) {
			this.children.add(new Concept(child));
		}

		this.properties = new HashMap<String, Object>();
		for (String key : other.properties.keySet()) {
			this.properties.put(key, other.properties.get(key));
		}

		this.references = new ArrayList<LexikonEntry>(other.references);
	}

	public String getName() {
		return name;
	}

	public boolean isStructure() {
		return isStructure;
	}

	public void setStructure(boolean isStructure) {
		this.isStructure = isStructure;
	}

	public void setName(String name) {
		this.name = name;
	}

	public Set<Variation> getSynonyms() {
		return synonyms;
	}

	public void setSynonyms(Set<Variation> synonyms) {
		this.synonyms = synonyms;
	}

	public LexikonEntry getContext() {
		return context;
	}

	public void setContext(LexikonEntry context) {
		this.context = context;
	}

	@XmlTransient
	public Concept getParent() {
		return parent;
	}

	public void setParent(Concept parent) {
		this.parent = parent;
	}

	public List<Concept> getChildren() {

		if (this.children == null)
			children = new ArrayList<Concept>();
		return children;
	}

	public void setChildren(List<Concept> children) {
		this.children = children;
	}

	public HashMap<String, Object> getProperties() {
		return properties;
	}

	public void setProperties(HashMap<String, Object> properties) {
		this.properties = properties;
	}

	public void addSynonym(Variation synonym) {
		this.synonyms.add(synonym);
	}

	public void removeSynonym(String synonym) {
		this.synonyms.remove(synonym);
	}

	public void setProperty(String property, Object value) {

		this.properties.put(property, value);
	}

	public Object getProperty(String property) {
		return this.properties.get(property);
	}

	public boolean isRoot() {
		return isRoot;
	}

	public void setRoot(boolean isRoot) {
		this.isRoot = isRoot;
	}

	public boolean isAttribute() {

		if (this.userSetAttribute) {
			return this.isAttribute;
		}

		if (this.isRoot)
			return false;
		Concept parentTemp = parent;
		while (parentTemp.isStructure && parentTemp != null) {
			parentTemp = parentTemp.getParent();
		}
		if (parentTemp != null && parentTemp.isRoot) {
			return true;
		}
		return false;
	}

	public boolean isValue() {

		if (userSetValue) {
			return this.isValue;
		}

		if (parent == null)
			return false;
		Concept parentTemp = parent;
		while (parentTemp != null && parentTemp.isStructure) {
			parentTemp = parentTemp.getParent();
		}

		// get next step above
		parentTemp = parentTemp.getParent();
		while (parentTemp != null && parentTemp.isStructure) {
			parentTemp = parentTemp.getParent();
		}

		if (parentTemp != null && parentTemp.isRoot)
			return true;
		return false;
	}

	public List<Concept> getSuccessors() {

		List<Concept> nodes = new ArrayList<Concept>();

		if (children != null) {
			nodes.addAll(children);
			for (Concept c : children) {
				nodes.addAll(c.getSuccessors());
			}
		}

		return nodes;
	}

	public String toText(String input, int lvl) {

		// we add a tab and then our children

		for (int i = 0; i < lvl; i++) {
			input += "\t";
		}
		input += name + "  " + this.nodeType.toString().substring(0, 1) + "-";
		input += isAttribute() ? "A" : "V";
		if (isPureExistence)
			input += "-P";

		input += " ";

		input += "[";
		for (Variation v : this.synonyms) {

			if (v.getSynonym().equals(this.name))
				continue;
			input += v.getSynonym() + " " + v.getCategoriesAsString();
		}

		input += " LEX: ";
		for (LexikonEntry e : this.references) {

			for (Variation v : e.getAllVariations()) {

				if (v.getSynonym().equals(this.name))
					continue;
				input += v.getSynonym() + " " + v.getCategoriesAsString();
			}
		}

		input += "]";

		input += "\n";

		int lvlTemp = lvl + 1;
		for (Concept child : children) {
			String text = child.toText("", lvlTemp);
			input += text;
		}
		return input;
	}

	public boolean hasChild(Concept concept) {

		if (children == null)
			return false;
		for (Concept c : children) {

			if (c.equals(concept)) {
				return true;
			}
		}
		return false;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + id;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Concept other = (Concept) obj;
		if (id != other.id)
			return false;
		return true;
	}

	public int getId() {
		return id;
	}

	public void setId(int id) {
		this.id = id;
	}

	public void removeChild(Concept c) {

		children.remove(c);

	}

	public void addChild(Concept child) {

		if (this.children == null) {
			this.children = new ArrayList<Concept>();
		}

		this.children.add(child);

	}

	public List<String> getSynonymsAsRegex() {

		List<String> regexes = new ArrayList<String>();

		for (Variation s : synonyms) {

			// should exclude UNITS
			if (s.isUnit())
				continue;

			regexes.addAll(s.getVariationsAsList());

		}

		// also add all template entries

		for (LexikonEntry entry : this.references) {
			if (entry == null) {
				continue;
			}
			for (Variation v : entry.getVariations()) {
				regexes.addAll(v.getVariationsAsList());
			}
		}
		regexes.add("\\b" + Pattern.quote(name) + "\\b");
		return regexes;
	}

	public ENodeType getNodeType() {

		if (nodeType == null) {
			return ENodeType.DEFAULT;
		}
		return nodeType;
	}

	public void setNodeType(ENodeType nodeType) {
		this.nodeType = nodeType;
	}

	public boolean isImplicit() {
		return isImplicit;
	}

	public void setImplicit(boolean isImplicit) {
		this.isImplicit = isImplicit;
	}

	public boolean isObjectAttribute() {
		return isObjectAttribute;
	}

	public void setObjectAttribute(boolean isObjectAttribute) {
		this.isObjectAttribute = isObjectAttribute;
	}

	// used in the UI to dynamically display all options one can set of a
	// concept
	public static String[] getBooleanAttributesAsStringArray() {

		String[] attributes = new String[] { IS_ATTRIBUTE, IS_IMPLICIT, IS_OBJECT_ATTRIBUTE, IS_STRUCTURE, IS_VALUE,
				PURE_EXISTENCE };

		return attributes;
	}

	public void changeAttribute(String data) {

		if (data.equals(IS_OBJECT_ATTRIBUTE)) {
			this.isObjectAttribute = !this.isObjectAttribute;
		}

		if (data.equals(IS_ATTRIBUTE)) {
			this.isAttribute = !this.isAttribute;
			this.userSetAttribute = true;
		}

		if (data.equals(IS_IMPLICIT)) {
			this.isImplicit = !this.isImplicit;
		}

		if (data.equals(IS_STRUCTURE)) {
			this.isStructure = !this.isStructure;
		}

		if (data.equals(PURE_EXISTENCE)) {
			this.isPureExistence = !this.isPureExistence;
		}

		if (data.equals(IS_VALUE)) {
			this.isValue = !this.isValue;
			this.userSetValue = true;
		}

	}

	public boolean getAttribute(String data) {
		if (data.equals(IS_OBJECT_ATTRIBUTE)) {
			return this.isObjectAttribute;
		}

		if (data.equals(IS_ATTRIBUTE)) {
			return this.isAttribute();
		}

		if (data.equals(IS_IMPLICIT)) {
			return this.isImplicit;
		}

		if (data.equals(IS_STRUCTURE)) {
			return this.isStructure;
		}

		if (data.equals(IS_VALUE)) {
			return this.isValue();
		}

		if (data.equals(PURE_EXISTENCE)) {
			return this.isPureExistence;
		}

		return false;
	}

	// a unit is stored as a special synonym
	public List<String> getUnits() {

		List<String> units = new ArrayList<String>();
		for (Variation s : synonyms) {

			if (s.isUnit())
				units.add(s.getSynonym());
		}
		return units;
	}

	public void removeVariation(Variation v) {

		if (v.isLexikon()) {
			System.out.println(v.getEntry());

			List<LexikonEntry> toRemove = new ArrayList<LexikonEntry>();

			for (LexikonEntry entry : this.references) {
				if (entry.getName().equals(v.getEntry())) {
					toRemove.add(entry);
				}
			}

			this.references.removeAll(toRemove);
		} else {
			this.synonyms.remove(v);
		}
	}

	public List<LexikonEntry> getReferences() {
		return references;
	}

	public void setReferences(List<LexikonEntry> references) {
		this.references = references;
	}

	public void removeReference(LexikonEntry ref) {
		this.references.remove(ref);
	}

	public void addReference(LexikonEntry ref) {
		this.references.add(ref);
	}

	public List<Variation> getAllVariations() {

		List<Variation> synonyms2 = new ArrayList<Variation>(synonyms);

		for (LexikonEntry entry : this.references) {
			synonyms2.addAll(entry.getAllVariations());
		}

		// add the name itself explicitly
		synonyms2.add(new Variation(name));
		return synonyms2;
	}

	public void addVariation(Variation v) {
		addSynonym(v);

	}

	public boolean isPureExistence() {
		return isPureExistence;
	}

	public void setPureExistence(boolean isPureExistence) {
		this.isPureExistence = isPureExistence;
	}

}
